import unittest
import ipaddress
import json
from zof.objectview import ObjectView, to_json, from_json


def _test_dict():
    return dict(a=1, b=2, z=dict(y=9))


class ObjectViewTestCase(unittest.TestCase):
    def setUp(self):
        self.obj = ObjectView(_test_dict())

    def test_init(self):
        self.assertEqual(self.obj.__dict__, _test_dict())

    def test_property(self):
        self.assertEqual(self.obj.a, 1)
        self.assertEqual(self.obj.b, 2)
        self.assertEqual(self.obj.z, {'y': 9})
        with self.assertRaisesRegex(AttributeError, 'c'):
            print(self.obj.c)
        # Nested dictionaries aren't wrapped.
        with self.assertRaisesRegex(AttributeError,
                                    "'dict' object has no attribute 'y'"):
            print(self.obj.z.y)

    def test_special_property(self):
        obj = ObjectView({'get': 42, 'to_json': 'foo'})
        self.assertEqual(obj['get'], 42)
        self.assertEqual(obj.get, 42)

        self.assertEqual(obj.to_json, 'foo')
        obj.to_json = 4
        self.assertEqual(obj['to_json'], 4)
        self.assertEqual(obj.to_json, 4)
        del obj.to_json
        with self.assertRaises(AttributeError):
            print(obj.to_json)

    def test_setproperty(self):
        self.assertEqual(self.obj.a, 1)
        self.obj.a = 5
        self.assertEqual(self.obj.a, 5)
        self.obj.a = 1
        self.assertEqual(self.obj.a, 1)
        self.obj.c = 4
        self.assertEqual(self.obj.c, 4)
        del self.obj['c']
        with self.assertRaisesRegex(AttributeError, 'c'):
            print(self.obj.c)

    def test_contains(self):
        self.assertTrue('a' in self.obj)
        self.assertTrue('b' in self.obj)
        self.assertTrue('c' not in self.obj)
        self.assertFalse('zz' in self.obj)
        with self.assertRaisesRegex(TypeError,
                                    'attribute name must be string'):
            self.assertFalse(1 in self.obj)

    def test_len(self):
        self.assertEqual(len(self.obj), 3)

    def test_bool(self):
        self.assertTrue(self.obj)
        empty = ObjectView({})
        self.assertTrue(empty)
        self.assertEqual(len(empty), 0)

    def test_iteration(self):
        count = 0
        for key in self.obj:
            self.assertIn(key, self.obj)
            self.assertIn(key, ['a', 'b', 'z'])
            count += 1
        self.assertEqual(count, 3)

    def test_getitem(self):
        self.assertEqual(self.obj['a'], 1)
        self.assertEqual(self.obj['b'], 2)
        with self.assertRaisesRegex(KeyError, 'c'):
            print(self.obj['c'])

    def test_setitem(self):
        self.assertEqual(self.obj['a'], 1)
        self.obj['a'] = 6
        self.assertEqual(self.obj['a'], 6)
        self.obj['a'] = 1
        self.assertEqual(self.obj['a'], 1)
        self.obj['d'] = 4
        self.assertEqual(self.obj['d'], 4)
        del self.obj['d']
        with self.assertRaisesRegex(KeyError, 'd'):
            print(self.obj['d'])

    def test_delitem(self):
        with self.assertRaisesRegex(AttributeError, 'aa'):
            print(self.obj.aa)
        self.obj.aa = 9
        self.assertEqual(self.obj.aa, 9)
        del self.obj.aa
        with self.assertRaisesRegex(AttributeError, 'aa'):
            print(self.obj.aa)
        with self.assertRaisesRegex(AttributeError, 'bb'):
            del self.obj.bb
        with self.assertRaisesRegex(KeyError, 'dd'):
            del self.obj['dd']

    def test_callable(self):
        self.assertEqual(self.obj('a'), 1)
        self.assertEqual(self.obj('b'), 2)
        self.assertEqual(self.obj('c'), None)
        self.assertEqual(self.obj('c', default=5), 5)

    def test_eq(self):
        self.assertEqual(self.obj, _test_dict())
        self.assertTrue(self.obj == _test_dict())
        self.assertTrue(_test_dict() == self.obj)
        self.assertEqual(self.obj, ObjectView(_test_dict()))
        self.assertTrue(self.obj == ObjectView(_test_dict()))

    def test_neq(self):
        self.assertNotEqual(self.obj, {})
        self.assertFalse(self.obj != _test_dict())
        self.assertFalse(_test_dict() != self.obj)
        self.assertNotEqual(self.obj, ObjectView({}))
        self.assertFalse(self.obj != ObjectView(_test_dict()))

    def test_str(self):
        obj = ObjectView(dict(q=[1, 2, 3]))
        self.assertEqual(str(obj), '{"q":[1,2,3]}')

    def test_repr(self):
        obj = ObjectView(dict(q=[1, 2, 3]))
        self.assertEqual(repr(obj), "{'q': [1, 2, 3]}")

    def test_to_json(self):
        obj = ObjectView(dict(q=[1, 2, 3]))
        self.assertEqual(to_json(obj), '{"q":[1,2,3]}')

    def test_from_json(self):
        obj = from_json(b'{"a":1}')
        self.assertEqual(obj, {'a': 1})

    def test_ip_address(self):
        obj = ObjectView(
            dict(a=[
                ipaddress.ip_address('10.1.2.3'), ipaddress.ip_address(
                    '2001::1')
            ]))
        self.assertEqual(to_json(obj), '{"a":["10.1.2.3","2001::1"]}')

    def test_ip_address_repr(self):
        # Using repr on an ObjectView containing objects does not
        # yield valid YAML.
        obj = ObjectView(
            dict(a=[
                ipaddress.ip_address('10.1.2.3'), ipaddress.ip_address(
                    '2001::1')
            ]))
        self.assertEqual(
            repr(obj),
            "{'a': [IPv4Address('10.1.2.3'), IPv6Address('2001::1')]}")

    def test_format(self):
        data = dict(x=dict(a=1))
        obj = ObjectView(data)
        self.assertEqual('{:2s}'.format(obj),
                         '{\n  "x": {\n    "a": 1\n  }\n}')
        with self.assertRaisesRegex(ValueError,
                                    'does not support format_spec'):
            '{:4d}'.format(obj)

    def test_std_json(self):
        data = dict(c=5)
        obj = ObjectView(data)
        with self.assertRaisesRegex(TypeError, 'is not JSON serializable'):
            # python's standard json module does not support objectview.
            json.dumps(obj)



"""Layer2 Demo App

- Implements reactive forwarding for a vlan-aware layer 2 switch.
- Ignores LLDP packets.
- Does not support loops.

"""

import zof
from zof.pktview import pktview_from_list

APP = zof.Application('layer2')

# The forwarding table is a dictionary that maps:
#   datapath_id -> { (eth_dst, vlan_vid) -> (out_port, time) }

APP.forwarding_table = {}


@APP.message('channel_up')
def channel_up(event):
    """Set up datapath when switch connects."""
    APP.logger.info('%s Connected from %s (%d ports, version %d)',
                    event['datapath_id'], event['endpoint'],
                    len(event['datapath']), event['version'])
    APP.logger.info('%s Remove all flows', event['datapath_id'])

    DELETE_FLOWS.send()
    BARRIER.send()
    TABLE_MISS_FLOW.send()


@APP.message('channel_down')
def channel_down(event):
    """Clean up when switch disconnects."""
    APP.logger.info('%s Disconnected', event['datapath_id'])
    APP.forwarding_table.pop(event['datapath_id'], None)


@APP.message('packet_in', eth_type=0x88cc)
def lldp_packet_in(_event):
    """Ignore lldp packets."""
    APP.logger.debug('lldp packet ignored')


@APP.message('packet_in')
def packet_in(event):
    """Handle incoming packets."""
    APP.logger.debug('packet in %r', event)
    datapath_id = event['datapath_id']
    msg = event['msg']
    time = event['time']
    data = msg['data']

    # Check for incomplete packet data.
    if len(data) < msg['total_len']:
        APP.logger.warning('Incomplete packet data: %r', event)
        return

    in_port = msg['in_port']
    pkt = msg['pkt']
    vlan_vid = pkt('vlan_vid', default=0)

    # Retrieve fwd_table for this datapath.
    fwd_table = APP.forwarding_table.setdefault(datapath_id, {})

    # Update fwd_table based on eth_src and in_port.
    if (pkt.eth_src, vlan_vid) not in fwd_table:
        APP.logger.info('%s Learn %s vlan %s on port %s', datapath_id,
                        pkt.eth_src, vlan_vid, in_port)
        fwd_table[(pkt.eth_src, vlan_vid)] = (in_port, time)

    # Lookup output port for eth_dst. If not found, set output port to 'ALL'.
    out_port, _ = fwd_table.get((pkt.eth_dst, vlan_vid), ('ALL', None))

    if out_port != 'ALL':
        APP.logger.info('%s Forward %s vlan %s to port %s', datapath_id,
                        pkt.eth_dst, vlan_vid, out_port)
        LEARN_MAC_FLOW.send(
            vlan_vid=vlan_vid, eth_dst=pkt.eth_dst, out_port=out_port)
        PACKET_OUT.send(out_port=out_port, data=data)

    else:
        # Send packet back out all ports (except the one it came in).
        APP.logger.info('%s Flood %s to %s vlan %s', datapath_id,
                        pkt.get_description(), pkt.eth_dst, vlan_vid)
        PACKET_FLOOD.send(in_port=in_port, data=data)


@APP.message('flow_removed')
def flow_removed(event):
    """Handle flow removed message."""
    datapath_id = event['datapath_id']
    match = pktview_from_list(event['msg']['match'])
    eth_dst = match.eth_dst
    vlan_vid = match.vlan_vid
    reason = event['msg']['reason']

    APP.logger.info('%s Remove %s vlan %s (%s)', datapath_id, eth_dst,
                    vlan_vid, reason)

    fwd_table = APP.forwarding_table.get(datapath_id)
    if fwd_table:
        fwd_table.pop((eth_dst, vlan_vid), None)


@APP.message(any)
def other_message(event):
    """Log ignored messages."""
    APP.logger.debug('Ignored message: %r', event)


DELETE_FLOWS = zof.compile('''
  # Delete flows in table 0.
  type: FLOW_MOD
  msg:
    command: DELETE
    table_id: 0
''')

BARRIER = zof.compile('''
  type: BARRIER_REQUEST
''')

TABLE_MISS_FLOW = zof.compile('''
  # Add permanent table miss flow entry to table 0
  type: FLOW_MOD
  msg:
    command: ADD
    table_id: 0
    priority: 0
    instructions:
      - instruction: APPLY_ACTIONS
        actions:
          - action: OUTPUT
            port_no: CONTROLLER
            max_len: NO_BUFFER
''')

LEARN_MAC_FLOW = zof.compile('''
  type: FLOW_MOD
  msg:
    table_id: 0
    command: ADD
    idle_timeout: 60
    hard_timeout: 120
    priority: 10
    buffer_id: NO_BUFFER
    flags: [ SEND_FLOW_REM ]
    match:
      - field: ETH_DST
        value: $eth_dst
      - field: VLAN_VID
        value: $vlan_vid
    instructions:
      - instruction: APPLY_ACTIONS
        actions:
          - action: OUTPUT
            port_no: $out_port
            max_len: MAX
''')

PACKET_OUT = zof.compile('''
  type: PACKET_OUT
  msg:
    actions:
      - action: OUTPUT
        port_no: $out_port
        max_len: MAX
    data: $data
''')

PACKET_FLOOD = zof.compile('''
  type: PACKET_OUT
  msg:
    in_port: $in_port
    actions:
      - action: OUTPUT
        port_no: ALL
        max_len: MAX
    data: $data
''')

if __name__ == '__main__':
    zof.run()



from .asynctestcase import AsyncTestCase
import asyncio
from zof.asyncmap import asyncmap


class AsyncMapTestCase(AsyncTestCase):
    async def test_asyncmap(self):
        parallel = 0
        max_parallel = 0

        async def _test(target):
            nonlocal parallel, max_parallel
            try:
                parallel += 1
                if parallel > max_parallel:
                    max_parallel = parallel
                await asyncio.sleep(0.05)
                if target == 3:
                    raise ValueError(target)
                return target
            finally:
                parallel -= 1

        targets = [1, 2, 3, 4, 5]
        for parallelism in (1, 2, 3, 4, 5, 6, 8):
            output = set()
            max_parallel = 0
            async for next_result in asyncmap(
                _test, targets, parallelism=parallelism):
                try:
                    assert next_result.done()
                    result = await next_result
                    output.add('+%s' % result)
                except ValueError as ex:
                    output.add('x%s' % ex)
            self.assertEqual(output, {'+2', '+1', 'x3', '+5', '+4'})
            self.assertEqual(max_parallel, min(parallelism, len(targets)))



import sys
if sys.version_info[:3] < (3, 5, 1):
    import platform
    raise NotImplementedError(
        'Python %s is not supported by the zof framework. Python 3.5.1 or later required.'
        % platform.python_version())

# pylint: disable=wrong-import-position, redefined-builtin
from .api_application import Application  # noqa: E402,F401
from .api_run import run  # noqa: E402,F401
from .api_args import common_args  # noqa: E402,F401
from .api_compile import compile  # noqa: E402,F401
from .api_functions import (get_apps, set_apps, get_datapaths, find_datapath,
                            find_port, post_event, ensure_future, connect,
                            close, get_connections, add_identity)

__version__ = '0.9.0'



"""Implements AsyncTestCase class."""

import asyncio
import inspect
import unittest

# Test timeout in seconds
_TIMEOUT = 10.0


# Method decorator for async tests.
def _wrap_async(func):
    """ Method decorator for async tests. """

    def _wrap(self):
        task = asyncio.wait_for(func(self), _TIMEOUT)
        AsyncTestCase.loop.run_until_complete(task)

    return _wrap


class AsyncTestCase(unittest.TestCase):
    """TestCase subclass that supports `async def` tests.

    Coroutine functions are automatically wrapped in a function that runs them
    using `AsyncTestCase.loop.run_until_complete`.

    To use this class, just create a subclass and write test methods as usual:

        class MyTestCase(AsyncTestCase):

            async def test_foo(self):
                await asyncio.sleep(5)
    """

    @classmethod
    def setUpClass(cls):
        AsyncTestCase.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(cls.loop)
        # Wrap test_ functions and setUp/tearDown if they are coroutine
        # functions.
        for key in dir(cls):
            if key.startswith('test_') or key in ('setUp', 'tearDown'):
                func = getattr(cls, key)
                if inspect.iscoroutinefunction(func):
                    setattr(cls, key, _wrap_async(func))

    @classmethod
    def tearDownClass(cls):
        AsyncTestCase.loop.close()



import ipaddress
from .objectview import ObjectView
from .ofctl import convert_from_ofctl

# Reserved payload field.
PAYLOAD = 'payload'


def pktview_alias(name, converter=(lambda x: x)):
    """Construct property that aliases specified attribute.

    converter function should be idempotent.
    """

    def _fget(self):
        try:
            return converter(self.__dict__[name])
        except KeyError:
            raise AttributeError('PktView object has no attribute "%s"' %
                                 name) from None

    def _fset(self, value):
        self.__dict__[name] = value

    def _fdel(self):
        try:
            del self.__dict__[name]
        except KeyError:
            raise AttributeError('PktView object has no attribute "%s"' %
                                 name) from None

    return property(fget=_fget, fset=_fset, fdel=_fdel)


def _description(pkt):
    """Return description of packet."""
    # pylint: disable=too-many-return-statements,too-many-branches
    eth_type = pkt('eth_type')
    if eth_type == 0x0806:
        arp_op = pkt('arp_op')
        if arp_op == 1:
            return 'ARP:REQ'
        if arp_op == 2:
            return 'ARP:REPLY'
        return 'ARP:%s' % arp_op
    if eth_type == 0x0800:
        ip_proto = pkt('ip_proto')
        if ip_proto == 1:
            return 'ICMPv4'
        if ip_proto == 2:
            return 'IGMPv4'
        if ip_proto == 6:
            return 'TCPv4'
        if ip_proto == 17:
            return 'UDPv4'
        return 'IPv4:%s' % ip_proto
    if eth_type == 0x86dd:
        ip_proto = pkt('ip_proto')
        if ip_proto == 58:
            return 'ICMPv6'
        if ip_proto == 6:
            return 'TCPv6'
        if ip_proto == 17:
            return 'UDPv6'
        return 'IPv6:%s' % ip_proto
    if eth_type == 0x88cc:
        return 'LLDP'
    return 'ETH:%s' % eth_type


class PktView(ObjectView):
    """Concrete class that represents a packet's header fields and payload.

    Use `make_pktview()` to construct a PktView object. The framework client may
    use a custom PktView subclass to add extra features.
    """

    # Alias some packet fields.
    ip_ttl = pktview_alias('nx_ip_ttl')
    hop_limit = pktview_alias('nx_ip_ttl')
    ipv6_nd_res = pktview_alias('x_ipv6_nd_res')

    @property
    def src(self):
        result = self('ipv4_src') or self('ipv6_src')
        if not result:
            raise AttributeError('src')
        return result

    @property
    def dst(self):
        result = self('ipv4_dst') or self('ipv6_dst')
        if not result:
            raise AttributeError('dst')
        return result

    def get(self, key, default=None):
        """Allows PktView to be treated as a Python dict.
        """
        return self.__dict__.get(key, default)

    def items(self):
        """Allows PktView to be treated as a Python dict.
        """
        return self.__dict__.items()

    def get_description(self):
        return _description(self)

    PROTO_FIELD = {
        'ETHERNET': 'eth_type',
        'ARP': 'arp_op',
        'IPV4': 'ipv4_src',
        'IPV6': 'ipv6_src',
        'ICMPV4': 'icmpv4_type',
        'ICMPV6': 'icmpv6_type'
    }

    def get_protocol(self, protocol):
        """Check if pkt is of specified type and return `self`.

        Return None if pkt does not match specified protocol name. This API is
        intended to be similar to RYU's. The implementation is not intended to
        exhaustively check every field; the oftr tool is reponsible for
        populating all necessary protocol fields.
        """
        return self if PktView.PROTO_FIELD[protocol.upper()] in self else None


def make_pktview(**kwds):
    """Construct a new PktView object."""
    return PktView(kwds)


def pktview_from_list(fields, *, slash_notation=False):
    """Construct a PktView object from a list of field objects.

    A field object may be an ObjectView or a dict.

    Args:
        fields (Seq[ObjectView|dict]): Sequence of fields.
        slash_notation (bool): If true, convert to "value/mask" notation.
    """
    if not isinstance(fields, (list, tuple)):
        raise ValueError('Expected list or tuple')

    pkt = make_pktview()
    for field in fields:
        key = field['field'].lower()
        if key == PAYLOAD:
            raise ValueError('Field "payload" is reserved')
        if 'mask' in field:
            value = (field['value'], field['mask'])
            if slash_notation:
                value = '%s/%s' % value
        else:
            value = field['value']
        pkt[key] = value
    return pkt


def pktview_to_list(pkt):
    """Convert a PktView object (or dict) into a list of fields."""
    if not isinstance(pkt, (dict, ObjectView)):
        raise ValueError('Expected a dict or ObjectView')

    return [_make_field(k, v) for k, v in _iter_items(pkt) if k != PAYLOAD]


def pktview_from_ofctl(ofctl, *, validate=False):
    """Convert an 'ofctl' dict to a PktView object.

    Args:
        ofctl (dict): ofctl match dictionary
        validate (bool): raise exception on unrecognized match field
    """
    if not isinstance(ofctl, dict):
        raise ValueError('Expected a dict')

    return PktView(convert_from_ofctl(ofctl, validate=validate))


def _make_field(name, value):
    fname = name.upper()
    value = convert_slash_notation(fname, value)

    if isinstance(value, tuple):
        if len(value) != 2:
            raise ValueError('len(tuple) != 2')
        return dict(field=fname, value=value[0], mask=value[1])

    return dict(field=fname, value=value)


def _iter_items(obj):
    if isinstance(obj, dict):
        return obj.items()
    return obj.__dict__.items()


def convert_slash_notation(fname, value):
    """Convert string value in slash notation into a tuple (addr, mask).
    """
    assert fname.isupper()
    if not isinstance(value, str) or '/' not in value:
        return value

    if fname not in _IP_FIELDS:
        return tuple(value.split('/', 1))

    try:
        # ip_interface handles IP prefix notation "/nn"
        addr = ipaddress.ip_interface(value)
        return (addr.ip, addr.netmask)
    except ValueError:
        if not fname.startswith('IPV6_'):
            raise

    # Handle generic slash notation for IPv6 addresses.
    addr = value.split('/', 1)
    return (ipaddress.IPv6Address(addr[0]), ipaddress.IPv6Address(addr[1]))


_IP_FIELDS = {'IPV4_SRC', 'IPV4_DST', 'IPV6_SRC', 'IPV6_DST', 'IPV6_ND_TARGET'}



import re
import os
import difflib
from glob import glob
from zof.connection import Connection
from .asynctestcase import AsyncTestCase

_UPDATE_GOLDEN = False

HERE = os.path.abspath(os.path.dirname(__file__))
COOKBOOK_DIR = os.path.join(HERE, '../doc/cookbook')

RECIPE = re.compile(r'(?ms)\n\[source,yaml\]\n----\n(.*?\n)----\n')

# Pass these args when launching oftr.
OFTR_ARGS = '--roundtrip --ofversion=4'


class CookbookTestCase(AsyncTestCase):
    """
    Verify that YAML code in the "cookbook" asciidoc is valid, and roundtrips to
    the expected canonical form.
    """

    async def setUp(self):
        oftr_options = {'subcmd': 'encode', 'args': OFTR_ARGS}
        self.conn = Connection(oftr_options=oftr_options)
        await self.conn.connect()

    async def tearDown(self):
        self.conn.close(write=True)
        return_code = await self.conn.disconnect()
        if return_code:
            raise Exception('oftr exited with return code %d' % return_code)

    async def test_cookbooks(self):
        for filename in glob(os.path.join(COOKBOOK_DIR, '*.adoc')):
            name = os.path.basename(filename)
            with self.subTest(name):
                await self.check_file(name)

    async def check_file(self, filename):
        result = []
        for item in _read_cookbook(filename):
            if _UPDATE_GOLDEN:
                print(item)
            self._write(item)
            result += ['---\n']
            result += _comment(item)
            result += await self._read_msg()

        golden_filename = 'cookbook-%s.yml' % filename
        if _UPDATE_GOLDEN:
            _save(result, golden_filename)
        else:
            diff = _compare(result, golden_filename)
            for line in diff:
                print(line, end='')
            self.assertFalse(diff)

    def _write(self, data):
        self.conn.write(bytes(data, 'utf-8'), delimiter=b'\n---\n')

    async def _read_msg(self):
        result = []
        while True:
            line = await self.conn.readline(b'\n')
            if not line or line == b'...':
                break
            if line == b'---' or not line.rstrip():
                continue
            result.append(line.decode('utf-8') + '\n')
        return result


def _read_cookbook(filename):
    with open(os.path.join(COOKBOOK_DIR, filename)) as input_:
        data = input_.read()

    for item in RECIPE.finditer(data):
        yield item.group(1)


def _comment(s):
    return ['# %s' % line for line in s.splitlines(keepends=True)]


def _compare(result, filename):
    with open(os.path.join(HERE, filename)) as input_:
        data = input_.readlines()

    return list(difflib.unified_diff(data, result, n=1))


def _save(result, filename):
    with open(os.path.join(HERE, filename), 'w') as output_:
        output_.writelines(result)



import zof
import asyncio
from zof.demo.simulator import APP as SIM

APP = zof.Application('zbench', exception_fatal=True)

APP.flowmod_count = 0
APP.packetout_count = 0
APP.packetin_count = 0


@APP.event('start')
async def start(event):
    await asyncio.sleep(1)
    APP.flowmod_count = 0
    run_zbench()
    while True:
        await asyncio.sleep(0.25)
        APP.logger.info('%d packetin, %d flowmod, %d packetout',
                        APP.packetin_count, APP.flowmod_count,
                        APP.packetout_count)
        #if APP.flowmod_count and APP.flowmod_count >= APP.packetin_count:
        #    break


@APP.message('flow_mod', datapath_id=None)
def flowmod(event):
    #APP.logger.info('flowmod: %r', event)
    APP.flowmod_count += 1


@APP.message('packet_out', datapath_id=None)
def packet_out(event):
    APP.packetout_count += 1


@APP.message('packet_in')
def packet_in(event):
    APP.packetin_count += 1


PACKET_IN = zof.compile('''
type:            PACKET_IN
msg:             
  buffer_id:       NO_BUFFER
  total_len:       0x002E
  in_port:         0x00000002
  in_phy_port:     0x00000002
  metadata:        0x0000000000000000
  reason:          APPLY_ACTION
  table_id:        0x03
  cookie:          0x000000005ADC15C0
  match:           
    - field:       IN_PORT
      value:       0x00000002
  data:            B21F7DA0EBDB9EF45FA6B15781000064080600010800060400029EF45FA6B1570A000002B21F7DA0EBDB0A000001
''')


def run_zbench():
    for conn_id in SIM.conn_to_sim:
        APP.logger.info('Send packet_ins to conn_id %s', conn_id)
        for _ in range(1000):
            PACKET_IN.send(conn_id=conn_id)


if __name__ == '__main__':
    zof.run()



import signal
import asyncio
import functools

DEFAULT_SIGNALS = ('SIGINT', 'SIGHUP', 'SIGTERM')


def _default_signal_handler(_signame):
    asyncio.get_event_loop().stop()


def run_server(*,
               signals=DEFAULT_SIGNALS,
               signal_handler=_default_signal_handler,
               pending_timeout=5.0,
               logger=None):
    """Run asyncio event loop for server.

    This function handles the boilerplate for running an async server:

      - handling signals
      - shutting down asynchronously
      - closing the loop

    Example::

        async def task():
            await asyncio.sleep(3)
            asyncio.get_event_loop().stop()

        asyncio.ensure_future(task())
        run_server()
    """
    loop = asyncio.get_event_loop()
    assert not loop.is_closed()

    # Install signal handler function(s).
    if signal_handler:
        for signame in signals:
            loop.add_signal_handler(
                getattr(signal, signame),
                functools.partial(signal_handler, signame))

    try:
        if logger:
            logger.debug('run_server started: %r',
                         asyncio.Task.all_tasks(loop))
        # Run loop until stopped with `loop.stop()`.
        loop.run_forever()

    except KeyboardInterrupt:
        pass

    finally:
        _shutdown_pending(loop, pending_timeout, logger)
        if hasattr(loop, 'shutdown_asyncgens'):
            if logger:
                logger.debug('run_server shutdown_asyncgens')
            loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()
        if logger:
            logger.debug('run_server stopped')


def _shutdown_pending(loop, pending_timeout, logger):
    """Give pending tasks a chance to exit cleanly.

    Try 3 times to finish our pending tasks. This gives stopping tasks
    a chance to create more async tasks during shutdown.
    """
    if logger:
        logger.debug('run_server: shutdown pending')
    for _ in range(3):
        if not _run_pending(loop, pending_timeout, logger):
            break
    if logger:
        tasks = [t for t in asyncio.Task.all_tasks(loop) if not t.done()]
        if tasks:
            logger.warning('run_server: shutdown completed: %d tasks:\n  %s',
                           len(tasks), '\n  '.join(repr(t) for t in tasks))
        else:
            logger.debug('run_server: shutdown completed')


def _run_pending(loop, pending_timeout, logger):
    """Run until pending tasks are complete.

    Return true if we still have more pending tasks.
    """
    pending = asyncio.Task.all_tasks(loop)
    if not pending:
        return False
    if logger:
        logger.debug('run_server: run_pending %r', pending)
    try:
        loop.run_until_complete(asyncio.wait(pending, timeout=pending_timeout))
    except RuntimeError as ex:
        # `run_until_complete` throws an exception if new async tasks are
        # started by the pending tasks *and* they are still running when the
        # original tasks complete. Return true when this happens.
        if str(ex) != 'Event loop stopped before Future completed.':
            raise
    return True



import asyncio
import re
import aiohttp
import aiohttp.web as web
from .objectview import to_json, from_json
from .endpoint import Endpoint

_VAR_REGEX = re.compile(r'^\{(\w+)\}$')
_LOG_FORMAT = '%a "%r" %s %b "%{Referrer}i" "%{User-Agent}i"'

ClientResponseError = aiohttp.ClientResponseError


class HttpServer:
    """Simple async web server.
    """

    def __init__(self, *, logger=None):
        self.endpoint = None
        self.logger = logger
        self.web_app = web.Application()
        self.web_handler = None
        self.web_server = None

    async def start(self, endpoint):
        self.endpoint = Endpoint(endpoint)
        loop = asyncio.get_event_loop()
        self.web_handler = self.web_app.make_handler(
            loop=loop, access_log=self.logger, access_log_format=_LOG_FORMAT)
        await self.web_app.startup()

        self.web_server = await loop.create_server(
            self.web_handler, self.endpoint.host, self.endpoint.port)

        if self.logger:
            self.logger.info('HttpServer: Start listening on %s',
                             self.endpoint)

    async def stop(self):
        self.web_server.close()
        await self.web_server.wait_closed()

        await self.web_app.shutdown()
        await self.web_handler.shutdown(timeout=10)
        await self.web_app.cleanup()

        if self.logger:
            self.logger.info('HttpServer: Stop listening on %s', self.endpoint)

    def get_json(self, path):
        route_path, route_vars = _split_route(path)

        def _wrap(func):
            get = _route_get_json(route_vars, func)
            self.web_app.router.add_get(route_path, get)

        return _wrap

    def post_json(self, path):
        route_path, route_vars = _split_route(path)

        def _wrap(func):
            post = _route_post_json(route_vars, func)
            self.web_app.router.add_post(route_path, post)

        return _wrap

    def get_text(self, path):
        route_path, route_vars = _split_route(path)

        def _wrap(func):
            get = _route_get_text(route_vars, func)
            self.web_app.router.add_get(route_path, get)
            return func

        return _wrap

    def post_text(self, path):
        route_path = route_vars = _split_route(path)

        def _wrap(func):
            post = _route_post_text(route_vars, func)
            self.web_app.router.add_post(route_path, post)

        return _wrap


def _split_route(path):
    """Split path on '?' into route_path and route_vars.

    For example, given "/path/{foo}?{bar}", return ('/path/{foo}', ['bar'])
    """
    if '?' not in path:
        return path, []
    route_path, rest = path.split('?', maxsplit=1)
    route_vars = []
    for name in rest.split('&'):
        m = _VAR_REGEX.match(name)
        if not m:
            raise ValueError("Invalid variable name: %s" % name)
        route_vars.append(m.group(1))
    return route_path, route_vars


def _route_get_json(route_vars, func):
    async def _get(request):
        kwds = _build_kwds(request, route_vars)
        return await _respond_json(func, kwds)

    return _get


def _route_get_text(route_vars, func):
    async def _get(request):
        kwds = _build_kwds(request, route_vars)
        return await _respond_text(func, kwds)

    return _get


def _route_post_json(route_vars, func):
    async def _post(request):
        post_data = await request.json(loads=from_json)
        kwds = _build_kwds(request, route_vars, post_data)
        return await _respond_json(func, kwds)

    return _post


def _route_post_text(route_vars, func):
    async def _post(request):
        post_data = await request.text()
        kwds = _build_kwds(request, route_vars, post_data)
        return await _respond_text(func, kwds)

    return _post


def _build_kwds(request, route_vars, post_data=None):
    kwds = request.match_info.copy()
    for var in route_vars:
        kwds[var] = request.query.get(var)
    if post_data is not None:
        kwds['post_data'] = post_data
    return kwds


async def _respond_json(func, kwds):
    if asyncio.iscoroutinefunction(func):
        result = await func(**kwds)
    else:
        result = func(**kwds)
    return web.json_response(result, dumps=to_json)


async def _respond_text(func, kwds):
    if asyncio.iscoroutinefunction(func):
        result = await func(**kwds)
    else:
        result = func(**kwds)
    # If result is a 2-tuple, treat it as (result, status)
    if isinstance(result, tuple):
        result, status = result
    else:
        status = 200
    if isinstance(result, bytes):
        return web.Response(
            body=result, status=status, content_type='text/plain')
    return web.Response(text=result, status=status)


class HttpClient:
    """Simple async web client.

    Usage:
        client = HttpClient()
        await client.start()

        result = await client.get_json('http://google.com')
        print(result)

        await client.stop()


    """

    def __init__(self):
        self._client = None

    async def start(self):
        assert self._client is None
        self._client = aiohttp.ClientSession(
            raise_for_status=True, json_serialize=to_json, conn_timeout=15)

    async def stop(self):
        # The following line requires aiohttp 2.2.2 or later.
        await self._client.close()
        self._client = None

    async def get_json(self, url):
        async with self._client.get(url) as response:
            return await response.json(loads=from_json)

    async def post_json(self, url, *, post_data):
        async with self._client.post(url, json=post_data) as response:
            return await response.json(loads=from_json)

    async def get_text(self, url):
        async with self._client.get(url) as response:
            return await response.text()

    async def post_text(self, url, *, post_data):
        async with self._client.post(url, text=post_data) as response:
            return await response.text()



import logging.config
import os
import warnings
import sys
import traceback

EXT_STDERR = 'ext://stderr'

DEFAULT_FORMATTER = logging.Formatter(
    '%(asctime)s.%(msecs)03d %(name)-6s %(levelname)-8s %(message)s',
    '%b %d %H:%M:%S')

STDERR_HANDLER = logging.StreamHandler(sys.stderr)
STDERR_HANDLER.setFormatter(DEFAULT_FORMATTER)

_LOGGING_INITED = False


def init_logging(loglevel, logfile=EXT_STDERR):
    """Set up logging.

    This method attaches log handlers to the root logger:
        - a stderr handler (only if the root logger has no other handlers)
        - a file handler if specified

    When a logfile is specified, the stderr handler will only log critical
    events.

    This routine enables asyncio debug mode if `loglevel` is 'debug'.
    """
    global _LOGGING_INITED  # pylint: disable=global-statement
    if _LOGGING_INITED:
        # Make sure we only initialize logging once.
        set_loglevel(loglevel)
        return
    _LOGGING_INITED = True

    _make_default_handlers(logfile)

    set_loglevel(loglevel)
    logging.captureWarnings(True)
    warnings.simplefilter('always')

    if loglevel.lower() == 'debug':
        # Make sure that asyncio debugging is enabled.
        os.environ['PYTHONASYNCIODEBUG'] = '1'

        # Log stack traces with all warnings.
        _formatwarning = warnings.formatwarning

        def _formatwarning_tb(*args, **kwargs):
            out = _formatwarning(*args, **kwargs)
            out += ''.join(traceback.format_stack()[:-1])
            return out

        warnings.formatwarning = _formatwarning_tb

    else:
        # When loglevel > debug, ignore ImportWarning's from uvloop.
        warnings.filterwarnings("ignore", category=ImportWarning)

    asyncio_logger = logging.getLogger('asyncio')
    asyncio_logger.setLevel('WARNING')


def set_loglevel(loglevel):
    """Change current log level.
    """
    ofp_logger = logging.getLogger(__package__)
    ofp_logger.setLevel(loglevel.upper())


def _make_default_handlers(logfile):
    """Prepare the default loggers.
    """
    root_logger = logging.getLogger()
    if not root_logger.hasHandlers():
        root_logger.addHandler(STDERR_HANDLER)

    if logfile and logfile != EXT_STDERR:
        logfile_handler = _make_logfile_handler(logfile)
        logfile_handler.setFormatter(DEFAULT_FORMATTER)
        root_logger.addHandler(logfile_handler)
        # When there is a log file, only log critical events to stderr.
        STDERR_HANDLER.setLevel('CRITICAL')


def _make_logfile_handler(logfile):
    """Return log file handler."""
    return logging.handlers.RotatingFileHandler(
        logfile, maxBytes=2**20, backupCount=10, encoding='utf8')



"""

Environment Variables:
    ZOF_OFTR_PREFIX
    ZOF_OFTR_PATH
    ZOF_OFTR_ARGS

"""

import argparse
import importlib
import sys
import os
from .controller import Controller
from .logging import EXT_STDERR

DEFAULT_ENDPOINTS = [6633, 6653]
DEFAULT_LOGFILE = EXT_STDERR
DEFAULT_LOGLEVEL = 'info'


class _ArgParserTest(argparse.ArgumentParser):
    """ArgumentParser subclass that does not exit.
    """

    def exit(self, status=0, message=None):
        raise RuntimeError('_ArgParserTest: status=%d, message="%s"' %
                           (status, message))


def _parent_args():
    return [
        app.arg_parser for app in Controller.singleton().apps
        if app.arg_parser is not None
    ]


def common_args(*, under_test=False, include_x_modules=False):
    """Construct default ArgumentParser parent.

    Args:
        under_test (Boolean): When true, create an argument parser subclass
            that raises an exception instead of calling exit.
        include_x_modules (Boolean): When true, include "--x-modules" argument.
    """
    if include_x_modules:
        _import_extra_modules()

    parser_class = _ArgParserTest if under_test else argparse.ArgumentParser
    parser = parser_class(
        add_help=False, parents=_parent_args(), conflict_handler='resolve')

    common_group = parser.add_argument_group('common arguments')
    common_group.add_argument(
        '-h', '--help', action='help', help='show this help message and exit')
    common_group.add_argument(
        '--loglevel',
        metavar='LEVEL',
        help='log level',
        default=DEFAULT_LOGLEVEL)
    common_group.add_argument(
        '--logfile', metavar='FILE', help='log file', default=DEFAULT_LOGFILE)
    common_group.add_argument('--pidfile', help='save pid file')

    listen_group = parser.add_argument_group('listen arguments')
    listen_group.add_argument(
        '--listen-endpoints',
        type=csv_list_type('endpoint'),
        metavar='ENDPOINT,...',
        help='listen endpoints separated by commas',
        default=DEFAULT_ENDPOINTS)
    listen_group.add_argument(
        '--listen-versions',
        type=csv_list_type(
            'version', item_type=int),
        metavar='VERSION,...',
        help='listen versions (1-6) separated by commas')
    listen_group.add_argument(
        '--listen-cert',
        type=file_contents_type(),
        metavar='FILE',
        help='certificate chain')
    listen_group.add_argument(
        '--listen-cacert',
        type=file_contents_type(),
        metavar='FILE',
        help='certificate authority')
    listen_group.add_argument(
        '--listen-privkey',
        type=file_contents_type(),
        metavar='FILE',
        help='private key')

    x_group = parser.add_argument_group('experimental')
    x_group.add_argument(
        '--x-oftr-path',
        help='path to oftr executable',
        default=os.getenv('ZOF_OFTR_PATH'))
    x_group.add_argument(
        '--x-oftr-args',
        help='arguments passed to oftr',
        default=os.getenv('ZOF_OFTR_ARGS'))
    x_group.add_argument(
        '--x-oftr-prefix',
        help='prefix used to launch oftr (valgrind, strace, catchsegv)',
        default=os.getenv('ZOF_OFTR_PREFIX'))
    x_group.add_argument(
        '--x-under-test',
        action='store_true',
        default=under_test,
        help='special test mode')

    # --x-modules is included here for documentation purposes only. This arg
    # is pre-processed by the `import_extra_modules()` function.
    if include_x_modules:
        x_group.add_argument(
            '--x-modules',
            type=csv_list_type('module'),
            metavar='MODULE,...',
            help='modules to import')

    # Various options to control performance experiments.
    xp_group = parser.add_argument_group('performance')
    xp_group.add_argument(
        '--xp-uvloop', action='store_true', help='use uvloop for asyncio')
    xp_group.add_argument(
        '--xp-ujson', action='store_true', help='use ujson for parsing JSON')
    xp_group.add_argument(
        '--xp-streams',
        action='store_true',
        help='use streams implementation (deprecated)')

    return parser


def _import_extra_modules(*, under_test=False):
    """Find the --x-modules argument and process it alone.
    """
    parser_class = _ArgParserTest if under_test else argparse.ArgumentParser
    parser = parser_class(add_help=False)
    parser.add_argument(
        '--x-modules',
        type=csv_list_type('module'),
        metavar='MODULE,...',
        help='modules to import')

    args, _ = parser.parse_known_args()
    if args.x_modules:
        _import_modules(args.x_modules)


def _import_modules(modules):
    """Import modules."""
    try:
        for module in modules:
            importlib.import_module(module)
    except ImportError as ex:
        print(ex, file=sys.stderr, flush=True)
        sys.exit(1)


def file_contents_type(name='file_contents_type', *, encoding='utf-8'):
    """Return contents of file."""

    def _parse(value):
        with open(value, encoding=encoding) as afile:
            return afile.read()

    _parse.__name__ = name
    return _parse


def csv_list_type(name='csv_list_type', *, item_type=str):
    """Return list of comma-separated values."""

    def _parse(value):
        return [item_type(s.strip()) for s in value.split(',')]

    _parse.__name__ = name
    return _parse



import asyncio


async def asyncmap(coro,
                   seq,
                   *,
                   ensure_future=asyncio.ensure_future,
                   parallelism=1):
    """
    Returns a coroutine object that, when awaited, yields the next result.

    This function lets you specify the amount of parallelism.

        async def request(target):
            ...

        targets = [1, 2, 3, 4, 5]

        async for next_result in async_map(request, targets, parallelism=2):
            try:
                # You MUST await the result even though it is done.
                assert next_result.done()
                result = await next_result
                ...
            except Exception:
                pass
    """

    assert seq
    pending = [ensure_future(coro(param)) for param in seq[0:parallelism]]
    next_idx = parallelism

    while pending:
        done, pending = await asyncio.wait(
            pending, return_when=asyncio.FIRST_COMPLETED)
        for fut in done:
            if next_idx < len(seq):
                pending.add(ensure_future(coro(seq[next_idx])))
                next_idx += 1
            yield fut



import unittest
import zof
import codecs
from zof.codec import encode, decode, OFTR_ENCODE, _oftr_call
from zof.exception import CodecError


class CodecTestCase(unittest.TestCase):
    def test_empty(self):
        with self.assertRaisesRegex(ValueError,
                                    r'YAML:1:1: error: not a document'):
            encode('')

        with self.assertRaisesRegex(ValueError,
                                    r'YAML:1:1: error: not a document'):
            encode('---\n\n...\n')

    def test_oftr_call(self):
        result = _oftr_call(OFTR_ENCODE, b'type: HELLO', 4)
        self.assertEqual(
            result,
            b'\x06\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00~')

    def test_hello(self):
        result = encode('type: HELLO\nversion: 1')
        self.assertEqual(result, b'\x01\x00\x00\x08\x00\x00\x00\x00')

        result = encode('type: HELLO')
        self.assertEqual(
            result,
            b'\x04\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00\x10')

        result = encode('type: HELLO', version=0)
        self.assertEqual(
            result,
            b'\x06\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00~')

    def test_hello_error(self):
        with self.assertRaisesRegex(
                ValueError, r"YAML:1:1: error: missing required key 'type'"):
            encode('typ: HELLO')

        with self.assertRaisesRegex(
                ValueError,
                r'YAML:1:7: error: unknown value "HELO" Did you mean "HELLO"?'):
            encode('type: HELO')

    def test_hello_object(self):
        result = encode({'type': 'HELLO', 'version': 4})
        self.assertEqual(
            result,
            b'\x04\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00\x10')

    def test_decode_hello(self):
        result = decode(b'\x01\x00\x00\x08\x00\x00\x00\x00')
        self.assertEqual(result, """---
type:            HELLO
xid:             0x00000000
version:         0x01
msg:             
  versions:        [  ]
...
""")

        result = decode(
            b'\x04\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00\x10')
        self.assertEqual(result, """---
type:            HELLO
xid:             0x00000000
version:         0x04
msg:             
  versions:        [ 4 ]
...
""")

        result = decode(
            b'\x06\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x08\x00\x00\x00~')
        self.assertEqual(result, """---
type:            HELLO
xid:             0x00000000
version:         0x06
msg:             
  versions:        [ 1, 2, 3, 4, 5, 6 ]
...
""")

    def test_codec(self):
        result = 'type: HELLO\nversion: 1'.encode('openflow')
        self.assertEqual(result, b'\x01\x00\x00\x08\x00\x00\x00\x00')

        result = b'\x01\x00\x00\x08\x00\x00\x00\x00'.decode('openflow')
        self.assertEqual(result, """---
type:            HELLO
xid:             0x00000000
version:         0x01
msg:             
  versions:        [  ]
...
""")

    def test_codec_exception(self):
        try:
            'type: PACKET_IN'.encode('openflow')
        except ValueError as ex:
            self.assertIsNone(ex.__cause__)
            self.assertIsInstance(ex, CodecError)

        try:
            codecs.encode('type: PACKET_IN', 'openflow')
        except ValueError as ex:
            self.assertIsNone(ex.__cause__)
            self.assertIsInstance(ex, CodecError)



import sys
import shlex
import inspect
import asyncio
import argparse
import logging
import collections

from prompt_toolkit.shortcuts import prompt_async
from prompt_toolkit.styles import style_from_dict
from prompt_toolkit.token import Token
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.contrib.completers import WordCompleter
from prompt_toolkit import AbortAction

import zof
from ..event import make_event
from ..exception import ControlFlowException
from ..logging import DEFAULT_FORMATTER, STDERR_HANDLER


class CommandException(ControlFlowException):
    """Exception class used by app commands to exit.

    This exception is caught by the command_shell service.
    """

    def __init__(self, *, status, message=None):
        super().__init__()
        self.status = status
        self.message = message

    def __str__(self):
        return '[CommandException status=%s, message=%s]' % (self.status,
                                                             self.message)


class _ArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        if asyncio.Task.current_task():
            raise CommandException(status=0)
        super().exit(status, message)


APP = zof.Application('command_shell')
APP.foreground_task = None
APP.command_prompt = '> '
APP.commands = {}

_Handler = collections.namedtuple('_Handler', 'subtype func options')


def _command(subtype, **kwds):
    """Command decorator.
    """

    def _wrap(func):
        upper_type = subtype.upper()
        APP.commands[upper_type] = _Handler(upper_type, func, kwds)

    return _wrap


APP.command = _command
APP.command.ArgumentParser = _ArgumentParser


@APP.event('start')
async def command_shell(_event):
    """Async task to listen for input and execute commands."""
    # The command shell task can be interrupted with CTRL-C (KeyboardInterrupt).
    APP.foreground_task = asyncio.Task.current_task()
    cmds = [h.subtype.lower() for h in all_command_handlers()]
    completer = WordCompleter(cmds)
    history = InMemoryHistory()
    bold_style = style_from_dict({Token: 'bold'})
    while True:
        try:
            command = await prompt_async(
                APP.command_prompt,
                history=history,
                completer=completer,
                style=bold_style,
                patch_stdout=True,
                complete_while_typing=False,
                on_abort=AbortAction.RETRY)
            if command:
                await run_command(command)
        except EOFError:
            zof.post_event('EXIT')
            break
    APP.foreground_task = None


@APP.event('signal', signal='SIGINT')
def handle_sigint(event):
    """Cancel async command_shell task."""
    if APP.foreground_task:
        APP.foreground_task.cancel()
        event.exit = False


async def run_command(command):
    """Run shell command."""
    cmd = shlex.split(command)
    handler = find_command_handler(cmd[0])
    if not handler:
        print('%s: command not found' % cmd[0])
    else:
        try:
            result = exec_command(cmd, handler)
            # Result of executing a handler may be:
            #  None:  handler already printed output.
            #  Iterable: we are responsible for printing output.
            #  Future: we should await the future.
            if result is None:
                return
            if inspect.isawaitable(result):
                await result
            else:
                for line in result:
                    print(line)
        except CommandException as ex:
            # Stay silent for commands that exit with status 0.
            if ex.status != 0:
                APP.logger.exception(ex)
        except asyncio.CancelledError:
            APP.logger.debug('CancelledError')
        except Exception as ex:  # pylint: disable=broad-except
            APP.logger.exception(ex)


def exec_command(cmd, handler):
    parser = handler.options.get('argparser')
    if parser:
        assert isinstance(parser, APP.command.ArgumentParser)
        # Make sure `prog` is set to correct command name.
        parser.prog = cmd[0]
        cmd = parser.parse_args(cmd[1:])
    event = make_event(event='COMMAND', args=cmd)
    return handler.func(event)


def find_command_handler(cmd):
    return APP.commands.get(cmd.upper())


def all_command_handlers():
    return list(APP.commands.values())


# =============== #
# C O M M A N D S #
# =============== #


def _help_brief(handler):
    brief = handler.options.get('brief')
    if brief:
        return brief
    parser = handler.options.get('argparser')
    if parser and parser.description:
        # Return first line of parser description.
        return parser.description.strip().split('\n', 1)[0]
    return 'No help.'


def _help_args():
    desc = '''List all commands or show help for a specific command.'''
    parser = APP.command.ArgumentParser(description=desc)
    parser.add_argument('command', nargs='?', help='name of command')
    return parser


@APP.command('help', argparser=_help_args())
def help_cmd(event):
    """List all commands or show help for a specific command."""
    cmd_name = event.args.command
    if cmd_name:
        yield _show_help(cmd_name)
    else:
        for handler in sorted(all_command_handlers()):
            yield '%-12s - %s' % (handler.subtype.lower(),
                                  _help_brief(handler))


def _show_help(cmd_name):
    handler = find_command_handler(cmd_name)
    if not handler:
        return '%s: command not found' % cmd_name
    parser = handler.options.get('argparser')
    if parser:
        parser.prog = cmd_name
        return parser.format_help()
    return ''


def _ls_args():
    parser = APP.command.ArgumentParser(
        description='List datapaths, ports or flows.')
    parser.add_argument(
        '-l', '--long', action='store_true', help='list in long format')
    return parser


@APP.command('ls', argparser=_ls_args())
def ls_cmd(event):
    """(TODO in device service) List datapaths, ports, or flows."""
    print(event)


def _ps_args():
    desc = '''List all running apps/tasks.

    Columns:
      PREC  app precedence value (empty for tasks)
      NAME  app name
    '''
    parser = APP.command.ArgumentParser(description=desc)
    parser.add_argument(
        '-a', '--all', action='store_true', help='show all tasks')
    return parser


@APP.command('ps', argparser=_ps_args())
def ps_cmd(event):
    """List all running apps/tasks."""
    app_tasks = collections.defaultdict(list)
    if event.args.all:
        for task in asyncio.Task.all_tasks():
            capp = getattr(task, 'zof_task_app', None)
            app_tasks[capp].append(task)

    yield 'PREC NAME'
    for capp in zof.get_apps():
        yield '%4d %s' % (capp.precedence, capp.name)
        for task in app_tasks[capp]:
            yield '       %s:%s' % (_task_name(task), task.zof_task_scope)
    for task in app_tasks[None]:
        yield '   - %s' % (_task_name(task))


def _task_name(task):
    # pylint: disable=protected-access
    return task._coro.__qualname__


def _log_args():
    desc = '''Show recent log messages.'''
    parser = APP.command.ArgumentParser(description=desc)
    parser.add_argument('-f', action='store_true', help='continue logging')
    return parser


@APP.command('log', argparser=_log_args())
async def log_cmd(event):
    """Show recent log messages."""
    if event.args.f:
        _CONSOLE_HANDLER.write('Type CTRL-C to stop logging.')
    # Print out lines from tail buffer.
    for line in _TAIL_HANDLER.lines():
        _CONSOLE_HANDLER.write(line)
    if event.args.f:
        # Temporarily change console log handler's level to use the root log level.
        console_level = _CONSOLE_HANDLER.level
        _CONSOLE_HANDLER.setLevel('NOTSET')
        try:
            # Wait for cancellation.
            while True:
                await asyncio.sleep(60)
        finally:
            # Change console log handler's level back.
            _CONSOLE_HANDLER.setLevel(console_level)


@APP.command('exit', brief='Exit command shell.')
def exit_cmd(_event):
    """Exit command shell."""
    zof.post_event('EXIT')


class TailBufferedHandler(logging.Handler):
    """Logging handler that records the last N log records."""

    def __init__(self, maxlen=10):
        super().__init__()
        self._tail = collections.deque(maxlen=maxlen)

    def lines(self):
        """Return last N log lines."""
        return self._tail

    def emit(self, record):
        """Log the specified log record."""
        try:
            self._tail.append(self.format(record))
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)

    def close(self):
        """Close the log handler."""
        super().close()
        self._tail.clear()

    @staticmethod
    def install():
        """Install tail logging handler."""
        handler = TailBufferedHandler()
        handler.setFormatter(DEFAULT_FORMATTER)
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        return handler


class PatchedConsoleHandler(logging.Handler):
    """Logging handler that writes to stdout EVEN when stdout is patched.

    The normal StreamHandler grabs a reference to `sys.stdout` once at
    initialization time. This class always logs to the current sys.stdout
    which may be patched at runtime by prompt_toolkit.

    This class disables the default StreamHandler if it is logging to stderr.
    """

    def emit(self, record):
        try:
            self.write(self.format(record))
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)

    def write(self, line):  # pylint: disable=no-self-use
        stream = sys.stdout
        stream.write(line)
        stream.write('\n')

    @staticmethod
    def install():
        """Install stdout logging handler.

        Change level of default stderr logging handler.
        """
        handler = PatchedConsoleHandler()
        handler.setFormatter(DEFAULT_FORMATTER)
        handler.setLevel('WARNING')
        root_logger = logging.getLogger()
        root_logger.addHandler(handler)
        # From now on, only log critical events to stderr.
        STDERR_HANDLER.setLevel('CRITICAL')
        return handler


_TAIL_HANDLER = TailBufferedHandler.install()
_CONSOLE_HANDLER = PatchedConsoleHandler.install()

if __name__ == '__main__':
    zof.run()



import asyncio
import zof
import argparse


def _arg_parser():
    parser = argparse.ArgumentParser(
        prog='conntest', description='Connection Closer', add_help=False)
    parser.add_argument(
        '--close-interval',
        type=int,
        default=1,
        help='Interval at which to close connections (seconds)')
    parser.add_argument(
        '--close-all', action='store_true', help='Close all connections')
    return parser


APP = zof.Application(
    'conntest', exception_fatal=True, arg_parser=_arg_parser())


@APP.event('start')
async def start(_):
    while True:
        await asyncio.sleep(APP.args.close_interval)
        # Obtain a list of all connections.
        conns = await zof.get_connections()
        for conn in conns:
            # If the connection has an associated datapath_id, close it.
            if conn['datapath_id'] or APP.args.close_all:
                APP.logger.info('close %d %s', conn['conn_id'],
                                conn['datapath_id'])
                count = await zof.close(conn_id=conn['conn_id'])
                assert count <= 1


if __name__ == '__main__':
    zof.run()



import zof


APP = zof.Application('autoresponder')


@APP.event('start')
def start(_):
    # Only set up the filter once for all datapaths.
    _SET_FILTER.send()


@APP.message('channel_up')
def channel_up(_):
    _DELETE_FLOWS.send()
    _DELETE_GROUPS.send()
    _BARRIER.send()
    _TABLE_MISS.send()
    _ARP_HUB.send()
    _ECHO_REPLY_GROUP.send()
    _ECHO_REPLY.send()



@APP.message('packet_in')
def packet_in(event):
    APP.logger.info('packet_in: %r', event['msg']['pkt'])


_SET_FILTER = zof.compile({
    'method': 'OFP.SET_FILTER',
    'params': [
        {
            'filter': 'icmp',
            'action': 'GENERIC_REPLY'
        }
    ]
})

_DELETE_FLOWS = zof.compile('''
    type: FLOW_MOD
    msg:
      command: DELETE
      table_id: ALL
''')

_TABLE_MISS = zof.compile('''
    type: FLOW_MOD
    msg:
      command: ADD
      table_id: 0
      instructions:
        - instruction: APPLY_ACTIONS
          actions:
            - action: OUTPUT
              port_no: CONTROLLER
''')

_ARP_HUB = zof.compile('''
    type: FLOW_MOD
    msg:
      command: ADD
      table_id: 0
      priority: 1
      match:
        - field: ETH_TYPE
          value: 0x0806   # ARP
      instructions:
        - instruction: APPLY_ACTIONS
          actions:
            - action: OUTPUT
              port_no: ALL
''')

_BARRIER = zof.compile('''
    type: BARRIER_REQUEST
''')

_DELETE_GROUPS = zof.compile('''
    type: GROUP_MOD
    msg:
      command: DELETE
      type: ALL
      group_id: ALL
      buckets: []
''')

_ECHO_REPLY_GROUP = zof.compile('''
    type: GROUP_MOD
    msg:
      command: ADD
      type: ALL
      group_id: 1000
      buckets:
        - actions:
          - action: NX_REG_MOVE
            src: IPV4_DST
            dst: PKT_REG0
          - action: NX_REG_MOVE
            src: IPV4_SRC
            dst: IPV4_DST
          - action: NX_REG_MOVE
            src: PKT_REG0[0:32]
            dst: IPV4_SRC
          - action: SET_FIELD
            field: ICMPV4_TYPE
            value: 0
          - action: NX_REG_MOVE
            src: ETH_SRC
            dst: PKT_REG0[0:48]
          - action: NX_REG_MOVE
            src: ETH_DST
            dst: ETH_SRC
          - action: NX_REG_MOVE
            src: PKT_REG0[0:48]
            dst: ETH_DST
          - action: OUTPUT
            port_no: IN_PORT
''')

_ECHO_REPLY = zof.compile('''
    type: FLOW_MOD
    msg:
      command: ADD
      table_id: 0
      priority: 2
      match: 
        - field: IPV4_DST
          value: 10.0.0.1
        - field: ICMPV4_TYPE
          value: 8
      instructions:
        - instruction: APPLY_ACTIONS
          actions:
            - action: GROUP
              group_id: 1000
''')


if __name__ == '__main__':
    zof.run()



import argparse
import asyncio
import zof
from zof.api_args import file_contents_type
import zof.exception as _exc


def _arg_parser():
    parser = argparse.ArgumentParser(
        prog='simulator', description='Simulator Demo', add_help=False)
    parser.add_argument(
        '--sim-endpoint',
        metavar='ENDPOINT',
        help='endpoint to connect to',
        default='127.0.0.1:6653')
    parser.add_argument(
        '--sim-count',
        type=int,
        default=10,
        help='Number of datapaths to simulate')
    parser.add_argument(
        '--sim-port-count',
        type=int,
        default=5,
        help='Number of ports per datapath')
    parser.add_argument(
        '--sim-timeout',
        type=float,
        default=0,
        help='Seconds to run simulation')
    parser.add_argument(
        '--sim-cert', type=file_contents_type(), help='Simulator certificate')
    parser.add_argument(
        '--sim-privkey',
        type=file_contents_type(),
        help='Simulator private key')
    parser.add_argument(
        '--sim-cacert',
        type=file_contents_type(),
        help='Simulator CA certificate')
    parser.add_argument(
        '--sim-reconnect',
        type=float,
        default=0,
        help='Reconnect retry interval (0=never)')
    return parser


APP = zof.Application(
    'simulator',
    exception_fatal=True,
    arg_parser=_arg_parser(),
    has_datapath_id=False)
APP.tls_id = 0
APP.conn_to_sim = {}
APP.connect_count = 0


async def _exit_timeout(timeout):
    await asyncio.sleep(timeout)
    zof.post_event({'event': 'SIM_TIMEOUT'})


@APP.event('sim_timeout')
def sim_timeout(_):
    exit_status = 20 if APP.connect_count < APP.args.sim_count else 0
    zof.post_event({'event': 'EXIT', 'exit_status': exit_status})


@APP.event('prestart')
async def prestart(_):
    if APP.args.sim_cert:
        APP.tls_id = await zof.add_identity(
            cert=APP.args.sim_cert,
            cacert=APP.args.sim_cacert,
            privkey=APP.args.sim_privkey)


@APP.event('start')
def start(_):
    if APP.args.sim_timeout:
        zof.ensure_future(_exit_timeout(APP.args.sim_timeout))
    for i in range(APP.args.sim_count):
        sim = Simulator(hex(i + 1))  # 'ff:ff:00:00:00:00:00:01')
        zof.ensure_future(sim.start())


@APP.message('channel_up')
@APP.message('flow_mod')
@APP.message('packet_out')
def ignore(_):
    return


@APP.message('features_request')
def features_request(event):
    APP.connect_count += 1
    APP.conn_to_sim[event['conn_id']].features_request(event)


@APP.message('barrier_request')
def barrier_request(event):
    APP.conn_to_sim[event['conn_id']].barrier_request(event)


@APP.message('request.port_desc')
def request_portdesc(event):
    APP.conn_to_sim[event['conn_id']].request_portdesc(event)


@APP.message('request.desc')
def request_desc(event):
    APP.conn_to_sim[event['conn_id']].request_desc(event)


@APP.message('request.port_stats')
def request_portstats(event):
    APP.conn_to_sim[event['conn_id']].request_portstats(event)


@APP.message('request.table_features')
def request_tablefeatures(event):
    APP.conn_to_sim[event['conn_id']].request_tablefeatures(event)


@APP.message('channel_down')
def channel_down(event):
    APP.conn_to_sim[event['conn_id']].channel_down()


@APP.message('role_request')
def role_request(event):
    APP.conn_to_sim[event['conn_id']].role_request()


@APP.message(any)
def other(event):
    APP.logger.info('Unhandled message: %r', event)
    raise ValueError('Unexpected message: %r' % event)


class Simulator(object):
    # pylint: disable = no-self-use

    def __init__(self, datapath_id):
        self.datapath_id = datapath_id
        self.conn_id = None

    async def start(self):
        try:
            self.conn_id = await zof.connect(
                APP.args.sim_endpoint, versions=[4], tls_id=APP.tls_id)
            APP.conn_to_sim[self.conn_id] = self
        except _exc.RPCException:
            self.channel_down()

    def channel_down(self):
        APP.conn_to_sim.pop(self.conn_id, None)
        # If there's a reconnect interval, then schedule the restart.
        if APP.args.sim_reconnect:
            zof.ensure_future(self._restart(APP.args.sim_reconnect))

    def features_request(self, event):
        msg = {
            'type': 'FEATURES_REPLY',
            'xid': event['xid'],
            'flags': ['NO_ALERT'],
            'msg': {
                'datapath_id': self.datapath_id,
                'n_buffers': 0,
                'n_tables': 32,
                'capabilities': [],
                'ports': self._portdescs() if event['version'] < 4 else []
            }
        }
        zof.compile(msg).send()

    def barrier_request(self, event):
        msg = {'type': 'BARRIER_REPLY', 'xid': event['xid']}
        zof.compile(msg).send()

    def request_portdesc(self, event):
        msg = {
            'type': 'REPLY.PORT_DESC',
            'xid': event['xid'],
            'msg': self._portdescs()
        }
        zof.compile(msg).send()

    def request_desc(self, event):  # pylint: disable=no-self-use
        msg = {
            'type': 'REPLY.DESC',
            'xid': event['xid'],
            'msg': {
                'hw_desc': 'sim hw_desc',
                'mfr_desc': 'sim mfr_desc',
                'sw_desc': 'sim sw_desc',
                'serial_num': 'sim serial_num',
                'dp_desc': 'sim dp_desc'
            }
        }
        zof.compile(msg).send()

    def request_portstats(self, event):
        msg = {
            'type': 'REPLY.PORT_STATS',
            'xid': event['xid'],
            'msg': self._portstats()
        }
        zof.compile(msg).send()

    def request_tablefeatures(self, event):
        # This code currently ignores the contents of the TableFeatures request.
        msg = {
            'type': 'REPLY.TABLE_FEATURES',
            'xid': event['xid'],
            'msg': self._tablefeatures()
        }
        zof.compile(msg).send()

    def _portdescs(self):
        return [self._portdesc(i + 1) for i in range(APP.args.sim_port_count)]

    def _portdesc(self, port_no):  # pylint: disable=no-self-use
        macaddr = '%12.12x' % port_no
        return {
            'port_no': port_no,
            'hw_addr': macaddr,
            'name': 'port %d' % port_no,
            'config': [],
            'state': [],
            'ethernet': {
                'curr': [],
                'advertised': [],
                'supported': [],
                'peer': [],
                'curr_speed': 0,
                'max_speed': 0
            }
        }

    def _portstats(self):
        return [self._portstat(i + 1) for i in range(APP.args.sim_port_count)]

    def _portstat(self, port_no):
        return {
            'port_no': port_no,
            'duration': 0,
            'rx_packets': 0,
            'tx_packets': 0,
            'rx_bytes': 0,
            'tx_bytes': 0,
            'rx_dropped': 0,
            'tx_dropped': 0,
            'rx_errors': 0,
            'tx_errors': 0,
            'ethernet': {
                'rx_frame_err': 0,
                'rx_over_err': 0,
                'rx_crc_err': 0,
                'collisions': 0
            },
            'properties': []
        }

    def _tablefeatures(self):
        # FIXME(bfish): > 20, breaking into multipart request not working?
        return [self._tablefeature(i) for i in range(20)]

    def _tablefeature(self, table_id):
        actions = [hex(i) for i in range(100)]
        tables = list(range(table_id + 1, 254))
        return {
            'table_id': table_id,
            'name': 'Table %d' % table_id,
            'metadata_match': 0,
            'metadata_write': 0,
            'config': [0],
            'max_entries': 1024,
            'instructions': actions,
            'next_tables': tables,
            'write_actions': actions,
            'apply_actions': actions,
            'match': actions,
            'wildcards': actions,
            'write_set_field': actions,
            'apply_set_field': actions
        }

    async def _restart(self, interval):
        APP.logger.info('_restart in %d seconds', interval)
        await asyncio.sleep(interval)
        await self.start()


def main():
    args = zof.common_args(include_x_modules=True)
    args.set_defaults(listen_endpoints=None)
    zof.run(args=args.parse_args())


if __name__ == '__main__':
    main()



"""A setuptools based setup module for zof.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import os
import re
from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))
README_PATH = os.path.join(HERE, 'README.rst')
VERSION_PATH = os.path.join(HERE, 'zof', '__init__.py')


def _get_description(path):
    with open(path, encoding='utf-8') as afile:
        return afile.read()


def _get_version(path):
    with open(path, encoding='utf-8') as afile:
        regex = re.compile(r"(?m)__version__\s*=\s*'(\d+\.\d+\.\d+)'")
        return regex.search(afile.read()).group(1)


setup(
    name='zof',
    packages=find_packages(exclude=['test']),
    version=_get_version(VERSION_PATH),
    license='MIT',

    description='OpenFlow App Framework',
    long_description=_get_description(README_PATH),
    keywords='openflow controller',

    # The project's main homepage and author.
    url='https://github.com/byllyfish/zof',
    author='William W. Fisher',
    author_email='william.w.fisher@gmail.com',

    # Dependencies
    install_requires=[
        # Imported by http submodule. Required for metrics demo.
        'aiohttp>=2.2.2',
        # Required for metrics demo.
        'prometheus_client'
    ],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Networking'
    ],

    zip_safe=True,
    test_suite='test'
)



import unittest
import zof.exception as _exc


class ExceptionTestCase(unittest.TestCase):
    def test_timeout(self):
        ex = _exc.TimeoutException(1, 5)
        self.assertEqual(ex.xid, 1)
        self.assertEqual(ex.timeout, 5)
        self.assertEqual(ex.message, '')
        self.assertEqual(str(ex), '[TimeoutException xid=1 timeout=5]')

    def test_rpc(self):
        event = dict(id=2, error={'message': 'm', 'code': 1000})
        ex = _exc.RPCException(event)
        self.assertEqual(ex.xid, 2)
        self.assertEqual(ex.message, 'm')
        self.assertEqual(ex.code, 1000)
        self.assertEqual(str(ex), '[RPCException xid=2 message=m]')

    def test_error(self):
        event = dict(xid=3, type='ERROR', msg={})
        ex = _exc.ErrorException(event)
        self.assertEqual(ex.xid, 3)
        self.assertIs(ex.event, event)
        self.assertRegex(str(ex), r'\[ErrorException xid=3 event=.+\]')

    def test_delivery(self):
        event = dict(xid=4, type='FLOW_MOD', msg={})
        ex = _exc.DeliveryException(event)
        self.assertEqual(ex.xid, 4)
        self.assertIs(ex.event, event)
        self.assertRegex(str(ex), r'\[DeliveryException xid=4 event=.+\]')



"""Implements the ObjectView class."""

import argparse
import json
from ipaddress import IPv4Address, IPv6Address


class ObjectView(object):
    """Wraps a Python dictionary so its attributes can be accessed/assigned
    using dot notation `obj.attr` or `obj['spacey attr']`.

    Examples::

        obj = ObjectView({'get': 42})
        assert obj.get == obj['get'] == 42
        assert obj('to_json', default='foo') == 'foo'
        obj.to_json = 1
        assert obj.to_json == 1

    If you access a value that does not exist, ObjectView raises a KeyError.
    Use callable syntax to access a value using a default if the key is not
    present, e.g. `obj('foo', default='bar')`.

    This class does _not_ recursively wrap sub-objects; it's designed to be
    used as a `json.loads` object_hook.

    This class does not define any non-dunder methods to avoid conflict
    with attribute names.

    This implementation uses getattr, setattr, and delattr instead of __dict__
    where possible, so subclasses can define alias properties (see PktView).

    TODO(bfish): Implement __copy__ and __deepcopy__? Implement __getstate__?
    """

    def __init__(self, d):
        self.__dict__ = d

    def __contains__(self, key):
        """Make sure that `in` and `not in` still work: `if key in obj:`"""
        return hasattr(self, key)

    def __len__(self):
        """Make sure len(obj) works."""
        return len(self.__dict__)

    def __bool__(self):
        """Make sure empty object is still true."""
        return True

    def __iter__(self):
        """Make sure iteration works."""
        return iter(self.__dict__)

    def __getitem__(self, key):
        """Allow read access using dictionary syntax `obj[key]`."""
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key) from None

    def __setitem__(self, key, value):
        """Allow write access using dictionary syntax `obj[key] = value`."""
        setattr(self, key, value)

    def __delitem__(self, key):
        """Allow properties to be deleted using dictionary syntax `del obj[key]`."""
        try:
            delattr(self, key)
        except AttributeError:
            raise KeyError(key) from None

    def __call__(self, key, *, default=None):
        """Allow read access using callable syntax `obj(key)`.

        Returns a default (None) if the key is not present.
        """
        try:
            return getattr(self, key)
        except AttributeError:
            return default

    def __eq__(self, obj):
        """Make sure that `==` works as expected."""
        try:
            return self.__dict__ == obj.__dict__
        except AttributeError:
            return self.__dict__ == obj

    def __ne__(self, obj):
        """Make sure that `!=` works as expected."""
        return not self.__eq__(obj)

    def __repr__(self):
        """Return the dictionary representation."""
        return repr(self.__dict__)

    def __str__(self):
        """Return compact JSON representation."""
        return to_json(self.__dict__)

    def __format__(self, format_spec):
        """Return formatted JSON representation.

        '4s' means JSON with 4-space indent.
        """
        if len(format_spec) == 2 and format_spec[1] == 's':
            indent = int(format_spec[0])
            return to_json_pretty(self, indent)
        raise ValueError('ObjectView does not support format_spec: %s' %
                         format_spec)


# ------------------------------------------------------------------------------
# JSON Utilities are defined here because of interdependency with ObjectView.
# ------------------------------------------------------------------------------


def from_json(text):
    """Parse text as json.
    """
    # If `text` is a byte string, decode it as utf-8.
    if isinstance(text, bytes):
        text = text.decode('utf-8')
    return json.loads(text)


def to_json(obj):
    """Return string with compact json representation of an object.
    """
    return json.dumps(
        obj,
        separators=(',', ':'),
        ensure_ascii=False,
        default=_json_serialize)


def to_json_pretty(obj, indent=4):
    """Return string with indented json representation of an object.
    """
    return json.dumps(
        obj, ensure_ascii=False, default=_json_serialize, indent=indent, sort_keys=True)


def _json_serialize(obj):
    if isinstance(obj, bytes):
        return obj.hex()
    if isinstance(obj, (ObjectView, argparse.Namespace)):
        return vars(obj)
    if isinstance(obj, (IPv4Address, IPv6Address)):
        return str(obj)
    try:
        return obj.__getstate__()
    except AttributeError:
        raise TypeError('Value "%s" of type %s is not JSON serializable' %
                        (repr(obj), type(obj)))



import asyncio
import sys

_PY36 = sys.version_info[0:2] >= (3, 6)

if _PY36:
    from .asyncmap36 import asyncmap  # pylint: disable=unused-import

else:

    class asyncmap:  # pylint: disable=invalid-name
        """
        Async generator object that is a port of the asyncmap36 generator.

        This is needed for Python 3.5 which doesn't support "yield inside a coroutine".
        """

        def __init__(self,
                     coro,
                     seq,
                     *,
                     ensure_future=asyncio.ensure_future,
                     parallelism=1):
            assert seq
            self.coro = coro
            self.seq = seq
            self.ensure_future = ensure_future
            self.pending = [
                ensure_future(coro(param)) for param in seq[0:parallelism]
            ]
            self.done = set()
            self.next_idx = parallelism

        def __aiter__(self):
            return self

        async def __anext__(self):
            if self.done:
                return self.done.pop()
            if not self.pending:
                raise StopAsyncIteration
            self.done, self.pending = await asyncio.wait(
                self.pending, return_when=asyncio.FIRST_COMPLETED)
            for _ in self.done:
                if self.next_idx < len(self.seq):
                    self.pending.add(
                        self.ensure_future(
                            self.coro(self.seq[self.next_idx])))
                    self.next_idx += 1
            return self.done.pop()



import os
from .controller import Controller
from .controllerapp import ControllerApp
from .logging import init_logging

# Enable logging before the first Application is created.
_DEBUG = os.environ.get('ZOF_DEBUG')
if _DEBUG:
    init_logging('info' if _DEBUG.lower() == 'info' else 'debug')


class Application(object):
    """The *zof.Application* class represents a controller *app*.

    Your app's code registers handlers for events via an *Application* instance.

    Args:
        name (str): Name of the app.
        precedence (int): Precedence for app event dispatch.
        exception_fatal (bool|str): If true, abort app when a handler raises
          an exception. When the value is a string, it's treated as the name of
          the exception logger `zof.<exc_log>`.
        arg_parser (argparse.ArgumentParser): App's argument parser.

    Attributes:
        name (str): App name.
        logger (Logger): App's logger.
    """

    def __init__(self,
                 name,
                 *,
                 controller=None,
                 exception_fatal=False,
                 precedence=100,
                 arg_parser=None,
                 has_datapath_id=True):
        if controller is None:
            controller = Controller.singleton()
        if controller.find_app(name):
            raise ValueError('App named "%s" already exists.' % name)

        # Construct internal ControllerApp object.
        app = ControllerApp(
            controller=controller,
            name=name,
            ref=self,
            exception_fatal=exception_fatal,
            precedence=precedence,
            arg_parser=arg_parser,
            has_datapath_id=has_datapath_id)

        self._app = app
        self.name = app.name
        self.logger = app.logger

    @property
    def args(self):
        return self._app.controller.args

    @property
    def precedence(self):
        return self._app.precedence

    @property
    def phase(self):
        return self._app.controller.phase

    @property
    def oftr_connection(self):
        return self._app.controller.conn

    @property
    def handlers(self):
        # TODO(bfish): Implement later...
        # return list(self._app.handlers)
        raise NotImplementedError('Not implemeted yet')

    # Decorators

    def message(self, subtype, **kwds):
        """Message decorator.
        """

        def _wrap(func):
            return self._app.register(func, 'message', subtype, kwds)

        return _wrap

    def event(self, subtype, **kwds):
        """Event decorator.
        """

        def _wrap(func):
            return self._app.register(func, 'event', subtype, kwds)

        return _wrap



import unittest
from zof.api_compile import CompiledString
from zof.objectview import ObjectView


class CompiledStringTestCase(unittest.TestCase):
    def test_init(self):
        msg = '''
        type: ECHO_REQUEST
        msg:
          data: DEADBEEF
        '''
        cmsg = CompiledString(None, msg)

        expected = '''\
method: OFP.SEND
params:
  xid: $xid
  datapath_id: $datapath_id
  conn_id: $conn_id
  type: ECHO_REQUEST
  msg:
    data: DEADBEEF'''

        actual = cmsg._template.template
        self.assertEqual(expected, actual)

        task_locals = dict(datapath_id='dpid_1', conn_id=7)

        expected = '''\
method: OFP.SEND
params:
  xid: 1000
  datapath_id: "dpid_1"
  conn_id: 7
  type: ECHO_REQUEST
  msg:
    data: DEADBEEF'''

        actual = cmsg._complete(dict(xid=1000), task_locals)
        self.assertEqual(expected, actual)

    def test_dict_var(self):
        template = '''
        type: $type_
        msg: $msg
        '''
        cmsg = CompiledString(None, template)
        task_locals = dict(datapath_id=None, conn_id=None)

        expected = '''\
method: OFP.SEND
params:
  xid: 1
  datapath_id: null
  conn_id: 13
  type: "ECHO_REQUEST"
  msg: {"data":"414243444546"}'''

        # Test with msg as a dictionary.
        msg = dict(data=b'ABCDEF')
        actual = cmsg._complete(
            dict(
                xid=1, conn_id=13, type_='ECHO_REQUEST', msg=msg),
            task_locals)
        self.assertEqual(expected, actual)

        # Test with msg as an ObjectView.
        msg = ObjectView(msg)
        actual = cmsg._complete(
            dict(
                xid=1, conn_id=13, type_='ECHO_REQUEST', msg=msg),
            task_locals)
        self.assertEqual(expected, actual)

    def test_missing_argument(self):
        msg = '''
        type: ECHO_REQUEST
        msg:
          data: DEADBEEF
        '''
        cmsg = CompiledString(None, msg)

        with self.assertRaises(ValueError):
            cmsg._complete(dict(), dict())

    def test_unknown_argument(self):
        msg = '''
        type: ECHO_REQUEST
        msg:
          data: DEADBEEF
        '''
        cmsg = CompiledString(None, msg)
        task_locals = dict(datapath_id='dpid_1', conn_id=7)

        with self.assertRaisesRegex(ValueError,
                                    'Unknown keyword argument "x"'):
            cmsg._complete(dict(x=5), task_locals)



import unittest
import zof


class CompileTestCase(unittest.TestCase):

    def test_compile_string(self):
        """Test that a YAML string compiles to an OF message.
        """
        ofmsg = zof.compile('type: HELLO')
        self.assertEqual('''\
<zof.CompiledString args=['conn_id', 'datapath_id', 'xid']>
method: OFP.SEND
params:
  xid: $xid
  datapath_id: $datapath_id
  conn_id: $conn_id
  type: HELLO
</zof.CompiledString>''', repr(ofmsg))

    def test_compile_object(self):
        """Test that an object compiles to an OF message.
        """
        ofmsg = zof.compile({'type': 'HELLO'})
        self.assertEqual('''\
<zof.CompiledObject>
{
    "type": "HELLO"
}
</zof.CompiledObject>''', repr(ofmsg))

    def test_compile_object_packetout(self):
        """Test that an object of type PACKET_IN/OUT compiles to an OF message.
        """
        ofmsg = zof.compile({
            'type': 'PACKET_OUT',
            'msg': {
                'pkt': {
                    'eth_type': 0x0800
                }
            }
        })
        self.assertEqual('''\
<zof.CompiledObject>
{
    "msg": {
        "_pkt": [
            {
                "field": "ETH_TYPE",
                "value": 2048
            }
        ]
    },
    "type": "PACKET_OUT"
}
</zof.CompiledObject>''', repr(ofmsg))

    def test_compile_rpc_object(self):
        """Test that an object compiles to an OF message.
        """
        rpc_msg = zof.compile({'method': 'OFP.DESCRIPTION'})
        self.assertEqual('''\
<zof.CompiledObjectRPC>
{
    "method": "OFP.DESCRIPTION"
}
</zof.CompiledObjectRPC>''', repr(rpc_msg))



import unittest
from zof.handler import make_handler

NO_HELP = 'No help available'


class HandlerTestCase(unittest.TestCase):
    def test_func(self):
        h = make_handler(func, 'message', 'FEATURES_REQUEST')
        self.assertTrue(h.verify())
        self.assertEqual('message', h.type)
        self.assertEqual('FEATURES_REQUEST', h.subtype)
        self.assertEqual('Brief line.', h.help_brief())
        self.assertEqual('Brief line.\n\nThis is a test func.', h.help())

        h = make_handler(func_nodoc, 'message', 'ECHO_REQUEST')
        self.assertTrue(h.verify())
        self.assertEqual('message', h.type)
        self.assertEqual('ECHO_REQUEST', h.subtype)
        self.assertEqual(NO_HELP, h.help_brief())
        self.assertEqual(NO_HELP, h.help())

    def test_async_func(self):
        h = make_handler(async_func, 'message', 'FEATURES_REQUEST')
        self.assertTrue(h.verify())
        self.assertEqual('message', h.type)
        self.assertEqual('FEATURES_REQUEST', h.subtype)
        self.assertEqual('Async func.', h.help_brief())
        self.assertEqual('Async func.\n\nThis is an async func.', h.help())

    def test_bad_func(self):
        h = make_handler(bad_func, 'message', 'FEATURES_REQUEST')
        self.assertFalse(h.verify())
        self.assertEqual('message', h.type)
        self.assertEqual('FEATURES_REQUEST', h.subtype)
        self.assertEqual(NO_HELP, h.help_brief())
        self.assertEqual(NO_HELP, h.help())

    def test_any_subtype(self):
        h = make_handler(func, 'message', any)
        self.assertTrue(h.verify())
        self.assertEqual('message', h.type)
        self.assertEqual(any, h.subtype)
        self.assertEqual('Brief line.', h.help_brief())
        self.assertEqual('Brief line.\n\nThis is a test func.', h.help())

    def test_bad_subtype_function(self):
        h = make_handler(func, 'message', bad_func)
        self.assertFalse(h.verify())

    def test_message_filter(self):
        h1 = make_handler(func, 'message', 'PACKET_IN', {'cookie': 123})
        self.assertTrue(h1.verify())

        h2 = make_handler(func, 'message', 'PACKET_IN', {'x': 9})
        self.assertTrue(h2.verify())

        evt = {
            'datapath_id': '00:00:00:00:00:00:00:01',
            'type': 'PACKET_IN',
            'msg': {
                'cookie': 124
            }
        }
        self.assertFalse(h1.match(evt))
        self.assertFalse(h2.match(evt))
        evt['msg']['cookie'] = 123
        self.assertTrue(h1.match(evt))
        self.assertFalse(h2.match(evt))
        evt['type'] = 'PACKET_OUT'
        self.assertFalse(h1.match(evt))
        self.assertFalse(h2.match(evt))

    def test_message_datapath(self):
        # Test datapath_id=None filter on message handler.
        h = make_handler(func, 'message', 'PACKET_OUT', {})
        self.assertTrue(h.verify())

        evt = {
            'datapath_id': '00:00:00:00:00:00:00:01',
            'type': 'PACKET_OUT',
            'msg': {
                'cookie': 124
            }
        }
        self.assertTrue(h.match(evt))
        evt['datapath_id'] = None
        self.assertTrue(h.match(evt))
        del evt['datapath_id']
        self.assertFalse(h.match(evt))  # only case where it makes a difference

    def test_message_no_datapath(self):
        # Test datapath_id=None filter on message handler.
        h = make_handler(func, 'message', 'PACKET_OUT', {'datapath_id': None})
        self.assertTrue(h.verify())

        evt = {
            'datapath_id': '00:00:00:00:00:00:00:01',
            'type': 'PACKET_OUT',
            'msg': {
                'cookie': 124
            }
        }
        self.assertFalse(h.match(evt))
        evt['datapath_id'] = None
        self.assertFalse(h.match(evt))
        del evt['datapath_id']
        self.assertTrue(h.match(evt))

    def test_event_filter(self):
        h1 = make_handler(func, 'event', 'SIGNAL', {'signal': 'SIGHUP'})
        self.assertTrue(h1.verify())

        h2 = make_handler(func, 'event', 'SIGNAL', {'x': 3})
        self.assertTrue(h2.verify())

        evt = {'event': 'SIGNAL', 'signal': 'SIGTERM'}
        self.assertFalse(h1.match(evt))
        self.assertFalse(h2.match(evt))
        evt['signal'] = 'SIGHUP'
        self.assertTrue(h1.match(evt))
        self.assertFalse(h2.match(evt))


def func(event):
    """
    Brief line.

    This is a test func.
    """
    pass


def func_nodoc(event):
    pass


async def async_func(event):
    """Async func.

    This is an async func.
    """
    pass


def bad_func():
    # Func doesn't have __doc__.
    pass
